import xbmc
from helper import pluginmenu, utils
from . import common, videos, genre, tag, studio, person, boxsets


class Movies:
    def __init__(self, EmbyServer, SQLs):
        self.EmbyServer = EmbyServer
        self.SQLs = SQLs
        self.GenreObject = genre.Genre(EmbyServer, self.SQLs)
        self.TagObject = tag.Tag(EmbyServer, self.SQLs)
        self.StudioObject = studio.Studio(EmbyServer, self.SQLs)
        self.PersonObject = person.Person(EmbyServer, self.SQLs)
        self.BoxSetObject = boxsets.BoxSets(EmbyServer, self.SQLs)
        SQLsSpecialFeatures = self.SQLs.copy()
        SQLsSpecialFeatures["video"] = None
        self.videos = videos.Videos(self.EmbyServer, SQLsSpecialFeatures)

    def change(self, Item):
        common.load_ExistingItem(Item, self.EmbyServer, self.SQLs["emby"], "Movie")

        if not common.verify_content(Item, "movie"):
            return False

        xbmc.log(f"EMBY.core.movies: Process Item: {Item['Name']}", 0) # DEBUG
        common.SwopMediaSources(Item)  # 3D
        common.set_trailer(Item, self.EmbyServer)
        common.set_RunTimeTicks(Item)
        common.get_streams(Item)
        common.set_chapters(Item, self.EmbyServer.ServerData['ServerId'])
        common.set_common(Item, self.EmbyServer.ServerData['ServerId'], False)
        Item['TagItems'].append({"LibraryId": Item["LibraryId"], "Type": "Tag", "Id": f"999999993{Item['LibraryId']}", "Name":  f"{Item['LibraryName']} (Library)", "Memo": "library"})
        common.set_MetaItems(Item, self.SQLs, self.GenreObject, self.EmbyServer, "Genre", "GenreItems")
        common.set_MetaItems(Item, self.SQLs, self.StudioObject, self.EmbyServer, "Studio", "Studios")
        common.set_MetaItems(Item, self.SQLs, self.TagObject, self.EmbyServer, "Tag", 'TagItems')
        common.set_people(Item, self.SQLs, self.PersonObject, self.EmbyServer)
        self.SQLs["emby"].add_streamdata(Item['Id'], Item['Streams'])

        if not common.get_file_path(Item, "movies", Item['LibraryId'], self.EmbyServer.ServerData['ServerId']):
            return False

        if Item['UpdateItem']:
            common.delete_ContentItemReferences(Item, self.SQLs, "movie")
        else:
            Item['KodiPathId'] = self.SQLs['video'].get_add_path(Item['Path'], "movies")
            Item['KodiItemId'] = self.SQLs["video"].create_movie_entry()
            Item['KodiFileId'] = self.SQLs["video"].create_entry_file()

        common.set_ContentItem(Item, self.SQLs, self.EmbyServer, "movie")
        common.set_Genre_links(Item['KodiItemId'], self.SQLs, "movie", Item["GenreItems"])
        common.set_Studio_links(Item['KodiItemId'], self.SQLs, "movie", Item["Studios"])
        common.set_Tag_links(Item['KodiItemId'], self.SQLs, "movie", Item["TagItems"])
        common.set_Actor_links(Item['KodiItemId'], self.SQLs, "movie", Item["CastItems"])
        common.set_Writer_links(Item['KodiItemId'], self.SQLs, "movie", Item["WritersItems"])
        common.set_Director_links(Item['KodiItemId'], self.SQLs, "movie", Item["DirectorsItems"])
        self.SQLs["video"].set_Favorite_Tag(Item['UserData']['IsFavorite'], Item['KodiItemId'], "movie")
        Item['Unique'] = self.SQLs["video"].add_uniqueids(Item['KodiItemId'], Item['ProviderIds'], "movie", 'imdb')
        Item['RatingId'] = self.SQLs["video"].add_ratings(Item['KodiItemId'], "movie", "default", Item['CommunityRating'])

        if not Item['ProductionLocations']:
            Item['ProductionLocations'].append(None)

        if Item['UpdateItem']:
            # Update Boxset
            for BoxSet in self.EmbyServer.API.get_Items(Item['ParentId'], ["BoxSet"], True, True, {'GroupItemsIntoCollections': True}):
                BoxSet['LibraryId'] = Item['LibraryId']
                self.BoxSetObject.change(BoxSet)

            self.SQLs["video"].update_movie(Item['KodiItemId'], Item['KodiFileId'], Item['Name'], Item['Overview'], Item['ShortOverview'], Item['Tagline'], Item['RatingId'], Item['Writers'], Item['KodiArtwork']['poster'], Item['Unique'], Item['SortName'], Item['RunTimeTicks'], Item['OfficialRating'], Item['Genre'], Item['Directors'], Item['OriginalTitle'], Item['Studio'], Item['Trailer'], Item['KodiArtwork']['fanart'].get('fanart', ""), Item['ProductionLocations'][0], Item['PremiereDate'], Item['UserData']['PlayCount'], Item['UserData']['LastPlayedDate'], None, Item['Filename'])
            self.SQLs["emby"].update_reference_movie_musicvideo(Item['Id'], "Movie", Item['UserData']['IsFavorite'], Item['PresentationUniqueKey'])
            xbmc.log(f"EMBY.core.movies: UPDATE [{Item['KodiPathId']} / {Item['KodiFileId']} / {Item['KodiItemId']}] {Item['Id']}: {Item['Name']}", 1) # LOGINFO
        else:
            self.SQLs["video"].add_movie(Item['KodiItemId'], Item['KodiFileId'], Item['Name'], Item['Overview'], Item['ShortOverview'], Item['Tagline'], Item['RatingId'], Item['Writers'], Item['KodiArtwork']['poster'], Item['Unique'], Item['SortName'], Item['RunTimeTicks'], Item['OfficialRating'], Item['Genre'], Item['Directors'], Item['OriginalTitle'], Item['Studio'], Item['Trailer'], Item['KodiArtwork']['fanart'].get('fanart', ""), Item['ProductionLocations'][0], Item['Path'], Item['KodiPathId'], Item['PremiereDate'], Item['Filename'], Item['DateCreated'], Item['UserData']['PlayCount'], Item['UserData']['LastPlayedDate'], None)
            self.SQLs["emby"].add_reference_movie_musicvideo(Item['Id'], Item['LibraryId'], "Movie", Item['KodiItemId'], Item['UserData']['IsFavorite'], Item['KodiFileId'], Item['PresentationUniqueKey'], Item['EmbyPath'], Item['KodiPathId'])
            xbmc.log(f"EMBY.core.movies: ADD [{Item['KodiPathId']} / {Item['KodiFileId']} / {Item['KodiItemId']}] {Item['Id']}: {Item['Name']}", 1) # LOGINFO

        if Item['CriticRating']:
            Item['CriticRating'] = float(Item['CriticRating'] / 10.0)
            self.SQLs["video"].add_ratings(Item['KodiItemId'], "movie", "tomatometerallcritics", Item['CriticRating'])

        # Add Special features
        if 'SpecialFeatureCount' in Item:
            if int(Item['SpecialFeatureCount']):
                SpecialFeatures = self.EmbyServer.API.get_specialfeatures(Item['Id'])

                for SF_Item in SpecialFeatures:
                    SF_Item.update({'ParentId': Item['Id'], "LibraryId": Item['LibraryId']})
                    self.videos.change(SF_Item)

        self.SQLs["emby"].add_multiversion(Item, "Movie", self.EmbyServer.API, self.SQLs)
        utils.FavoriteQueue.put(((Item['KodiArtwork']['thumb'], Item['UserData']['IsFavorite'], f"{Item['Path']}{Item['Filename']}", Item['Name'], "media", 0),))
        return not Item['UpdateItem']

    # This updates: Favorite, LastPlayedDate, Playcount, PlaybackPositionTicks
    def userdata(self, Item):
        if Item['PlayedPercentage'] and Item['PlayedPercentage']:
            RuntimeSeconds = int(Item['PlaybackPositionTicks'] / Item['PlayedPercentage'] / 100000)
        else:
            RuntimeSeconds = 0

        common.set_playstate(Item)
        self.SQLs["video"].set_Favorite_Tag(Item['IsFavorite'], Item['KodiItemId'], "movie")
        self.SQLs["video"].update_bookmark_playstate(Item['KodiFileId'], Item['PlayCount'], Item['LastPlayedDate'], Item['PlaybackPositionTicks'], RuntimeSeconds)
        self.SQLs["emby"].update_favourite(Item['IsFavorite'], Item['Id'], "Movie")
        self.set_favorite(Item['IsFavorite'], Item['KodiFileId'], Item['KodiItemId'])
        pluginmenu.reset_querycache("Movie")
        xbmc.log(f"EMBY.core.movies: New resume point {Item['Id']}: {Item['PlaybackPositionTicks']}", 0) # LOGDEBUG
        xbmc.log(f"EMBY.core.movies: USERDATA [{Item['KodiFileId']} / {Item['KodiItemId']}] {Item['Id']}", 1) # LOGINFO

    def remove(self, Item):
        if common.delete_ContentItem(Item, self.SQLs, "movie", "Movie"):
            self.set_favorite(False, Item['KodiFileId'], Item['KodiItemId'])
            self.SQLs["video"].delete_movie(Item['KodiItemId'], Item['KodiFileId'])
            xbmc.log(f"EMBY.core.movies: DELETE [{Item['KodiItemId']} / {Item['KodiFileId']}] {Item['Id']}", 1) # LOGINFO

            if not Item['LibraryId']:
                common.update_multiversion(self.SQLs["emby"], Item, "Movie")
        else:
            self.SQLs["video"].delete_links_tags(Item['KodiItemId'], "movie")

    def set_favorite(self, IsFavorite, KodiFileId, KodiItemId):
        FullPath, Thumbnail, Itemname = self.SQLs["video"].get_favoriteData(KodiFileId, KodiItemId, "movie")
        utils.FavoriteQueue.put(((Thumbnail, IsFavorite, FullPath, Itemname, "media", 0),))
